"use strict";
// The MIT License (MIT)
//
// Copyright (c) 2020 The Prometheus Authors
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
Object.defineProperty(exports, "__esModule", { value: true });
const codemirror_next_autocomplete_1 = require("@nexucis/codemirror-next-autocomplete");
const lezer_promql_1 = require("lezer-promql");
const fuzzy = require("fuzzy");
const MatchOp = 'MatchOp', BinOp = 'BinOp', BinaryExpr = 'BinaryExpr', FunctionIdentifier = 'FunctionIdentifier', AggregateOp = 'AggregateOp', MetricIdentifier = 'MetricIdentifier', Identifier = 'Identifier', GroupingLabels = 'GroupingLabels', GroupingLabel = 'GroupingLabel', LabelMatchers = 'LabelMatchers', LabelMatcher = 'LabelMatcher', LabelName = 'LabelName', VectorSelector = 'VectorSelector', StringLiteral = 'StringLiteral';
const autocompleteNode = {
    MatchOp: {
        labels: lezer_promql_1.promQLSyntax[MatchOp],
        type: '',
    },
    BinaryExpr: {
        labels: lezer_promql_1.promQLSyntax[BinOp],
        type: '',
    },
    FunctionIdentifier: {
        labels: lezer_promql_1.promQLSyntax[FunctionIdentifier],
        type: 'function',
    },
    AggregateOp: {
        labels: lezer_promql_1.promQLSyntax[AggregateOp],
        type: 'keyword',
    },
};
const snippets = [
    {
        keyword: 'sum(rate(&lt;input vector>[5m]))',
        snippet: 'sum(rate(${<input vector>}[5m]))',
    },
    {
        keyword: 'histogram_quantile(&lt;quantile>, sum by(le) (rate(&lt;histogram metric>[5m])))',
        snippet: 'histogram_quantile(${<quantile>}, sum by(le) (rate(${<histogram metric>}[5m])))',
    },
    {
        keyword: 'label_replace(&lt;input vector>, "&lt;dst>", "&lt;replacement>", "&lt;src>", "&lt;regex>")',
        snippet: 'label_replace(${<input vector>}, "${<dst>}", "${<replacement>}", "${<src>}", "${<regex>}")',
    },
];
// parsedSnippets shouldn't be modified. It's only there to not parse everytime the above list of snippet
const parsedSnippets = snippets.map((s) => ({
    label: s.name || s.keyword,
    apply: codemirror_next_autocomplete_1.snippet(s.snippet),
    score: 0,
}));
function escape(text) {
    return text.replace(/[&<>"']/g, (m) => {
        switch (m) {
            case '&':
                return '&amp;';
            case '<':
                return '&lt;';
            case '>':
                return '&gt;';
            case '"':
                return '&quot;';
            default:
                return '&#039;';
        }
    });
}
// HybridComplete is going to provide a full completion result with or without a remote prometheus.
class HybridComplete {
    constructor(prometheusClient, fuzzyPre, fuzzyPost) {
        this.prometheusClient = prometheusClient;
        this.fuzzyPre = fuzzyPre || '';
        this.fuzzyPost = fuzzyPost || '';
    }
    promQL(context) {
        var _a, _b, _c, _d, _e, _f, _g;
        const { state, pos } = context;
        const tree = state.tree.resolve(pos, -1);
        if (((_a = tree.parent) === null || _a === void 0 ? void 0 : _a.name) === MetricIdentifier && tree.name === Identifier) {
            // Here we cannot know if we have to autocomplete the metric_name, or the function or the aggregation.
            // So we will just autocomplete everything
            if (this.prometheusClient) {
                return this.prometheusClient.labelValues('__name__').then((metricNames) => {
                    const result = [{ labels: metricNames, type: 'constant' }];
                    return this.arrayToCompletionResult(result.concat(autocompleteNode[FunctionIdentifier], autocompleteNode[AggregateOp]), tree.start, pos, context, state, true);
                });
            }
            return this.arrayToCompletionResult([autocompleteNode[FunctionIdentifier]].concat(autocompleteNode[AggregateOp]), tree.start, pos, context, state, true);
        }
        if (tree.name === GroupingLabels || (((_b = tree.parent) === null || _b === void 0 ? void 0 : _b.name) === GroupingLabel && tree.name === LabelName)) {
            // In this case we are in the given situation:
            //      sum by ()
            // So we have to autocomplete any labelName
            return this.labelNames(tree, pos, context, state);
        }
        if (tree.name === LabelMatchers || (((_c = tree.parent) === null || _c === void 0 ? void 0 : _c.name) === LabelMatcher && tree.name === LabelName)) {
            // In that case we are in the given situation:
            //       metric_name{} or {}
            return this.autocompleteLabelNamesByMetric(tree, pos, context, state);
        }
        if (((_d = tree.parent) === null || _d === void 0 ? void 0 : _d.name) === LabelMatcher && tree.name === StringLiteral) {
            // In this case we are in the given situation:
            //      metric_name{labelName=""}
            // So we can autocomplete the labelValue
            return this.autocompleteLabelValue(tree.parent, tree, pos, context, state);
        }
        if (tree.name === MatchOp) {
            return this.arrayToCompletionResult([autocompleteNode[MatchOp]], tree.start, pos, context, state);
        }
        if (((_e = tree.parent) === null || _e === void 0 ? void 0 : _e.name) === BinaryExpr) {
            return this.arrayToCompletionResult([autocompleteNode[BinaryExpr]], tree.start, pos, context, state);
        }
        if (((_f = tree.parent) === null || _f === void 0 ? void 0 : _f.name) === FunctionIdentifier) {
            return this.arrayToCompletionResult([autocompleteNode[FunctionIdentifier]], tree.start, pos, context, state);
        }
        if (((_g = tree.parent) === null || _g === void 0 ? void 0 : _g.name) === AggregateOp) {
            return this.arrayToCompletionResult([autocompleteNode[AggregateOp]], tree.start, pos, context, state);
        }
        return null;
    }
    autocompleteLabelValue(parent, current, pos, context, state) {
        if (!this.prometheusClient) {
            return null;
        }
        // First get the labelName.
        // By definition it's the firstChild: https://github.com/promlabs/lezer-promql/blob/0ef65e196a8db6a989ff3877d57fd0447d70e971/src/promql.grammar#L250
        let labelName = '';
        if (this.prometheusClient && parent.firstChild && parent.firstChild.name === LabelName) {
            labelName = state.sliceDoc(parent.firstChild.start, parent.firstChild.end);
        }
        // then find the metricName if it exists
        const metricName = this.getMetricNameInVectorSelector(current, state);
        return this.prometheusClient.labelValues(labelName, metricName).then((labelValues) => {
            // +1 to avoid to remove the first quote.
            return this.arrayToCompletionResult([{ labels: labelValues, type: 'text' }], current.start + 1, pos, context, state);
        });
    }
    autocompleteLabelNamesByMetric(tree, pos, context, state) {
        return this.labelNames(tree, pos, context, state, this.getMetricNameInVectorSelector(tree, state));
    }
    getMetricNameInVectorSelector(tree, state) {
        // Find if there is a defined metric name. Should be used to autocomplete a labelValue or a labelName
        // First find the parent "VectorSelector" to be able to find then the subChild "MetricIdentifier" if it exists.
        let currentNode = tree;
        while (currentNode && currentNode.name !== VectorSelector) {
            currentNode = currentNode.parent;
        }
        if (!currentNode) {
            // Weird case that shouldn't happen, because "VectorSelector" is by definition the parent of the LabelMatchers.
            return '';
        }
        // By definition "MetricIdentifier" is necessary the first child if it exists
        if (!currentNode.firstChild || currentNode.firstChild.name !== MetricIdentifier) {
            // If it doesn't exist then we are in the given situation
            //     {}
            return '';
        }
        // Let's move forward to the next child.
        currentNode = currentNode.firstChild;
        // By definition the next child should be an "Identifier" which contains the metricName
        if (!currentNode.firstChild || currentNode.firstChild.name !== Identifier) {
            return '';
        }
        currentNode = currentNode.firstChild;
        return state.sliceDoc(currentNode.start, currentNode.end);
    }
    labelNames(tree, pos, context, state, metricName) {
        return !this.prometheusClient
            ? null
            : this.prometheusClient.labelNames(metricName).then((labelNames) => {
                // this case can happen when you are in empty bracket. Then you don't want to remove the first bracket
                return this.arrayToCompletionResult([
                    {
                        labels: labelNames,
                        type: 'constant',
                    },
                ], tree.name === GroupingLabels || tree.name === LabelMatchers ? tree.start + 1 : tree.start, pos, context, state);
            });
    }
    arrayToCompletionResult(data, from, to, context, state, includeSnippet = false) {
        const text = state.sliceDoc(from, to);
        const options = [];
        for (const completionList of data) {
            for (const label of completionList.labels) {
                const completionResult = this.filter(escape(label), text, label, context);
                if (completionResult) {
                    options.push(completionResult);
                }
            }
        }
        if (includeSnippet) {
            for (const s of parsedSnippets) {
                const completionResult = this.filter(s.label, text, s.apply, context);
                if (completionResult) {
                    options.push(completionResult);
                }
            }
        }
        return {
            from: from,
            to: to,
            options: options,
        };
    }
    filter(label, text, apply, context, type) {
        if (context.filterType !== codemirror_next_autocomplete_1.FilterType.Fuzzy) {
            // in case the fuzzy filtering is not used, then use the filter method coming from CMN
            let score;
            if ((score = context.filter(label, text, false))) {
                return { label: label, apply: apply, type: type, score: score };
            }
        }
        const fuzzyOption = { pre: this.fuzzyPre, post: this.fuzzyPost };
        const result = fuzzy.match(text, label, fuzzyOption);
        if (result) {
            return { label: result.rendered, apply: label, type: type, score: result.score };
        }
        return null;
    }
}
exports.HybridComplete = HybridComplete;
//# sourceMappingURL=hybrid.js.map